@@ -121,6 +121,10 @@ namespace libsemigroups {
         edge_index_type pos;
         State()             = default;
         State(State const&) = default;
+        State(State&&)      = default;
+        State& operator=(State const&) = default;
+        State& operator=(State&&) = default;
+
         State(node_index_type vv, edge_index_type ppos) : v(vv), pos(ppos) {}
 
         bool valid() const noexcept {
@@ -544,10 +548,17 @@ namespace libsemigroups {
       std::vector<size_t> _scratch;
       std::vector<size_t> _suffix_index;
     };
+
     namespace suffix_tree_helper {
 
       void add_words(SuffixTree& st, std::vector<word_type> const& words);
 
+      template <typename T>
+      void add_words(SuffixTree& st, T first, T last) {
+        for (auto it = first; it != last; ++it) {
+          st.add_word(*it);
+        }
+      }
     };  // namespace suffix_tree_helper
 
   }  // namespace detail
